/*
laserspotcam        A software to track a laser pointer on a wall.
Copyright (C) 2005  Raphael Assenat

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Author: Raphael Assenat <raph@raphnet.net>
*/
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include "SDL.h"

#include "webcam.h"
#include "SDL_v4l.h"
#include "filters.h"

#include "avgdiff.h"
#include "hotspots.h"
#include "daemonize.h"

// config variables
static char *cfg_v4l_dev = "/dev/video0";
static int cfg_display_raw = 0;
static int cfg_visual_mode = 0;
static int cfg_verbose = 0;
static int cfg_light_switch_skip = 8;
static int cfg_bg_mode = 0;
static int cfg_test_mode = 0;

// prototypes
void draw_cross(void);
void do_rotate_anim(char *filename_prefix, int fast_mode);
void do_panorama(char *filename_prefix, int fast_mode);
void draw_grid(int w, int h);
void draw_box(Uint32 color, int x, int y, int w, int h);
void draw_hotspots(void);
void waitKeyPress();
int calculateMotionBox(char *buf, int w, int h, SDL_Rect *r);
int findAndTriggerHotspot(SDL_Rect *mrect);
void tripleBuf(unsigned char *src, unsigned char *dst, int srclen);

int visual_main();
int daemon_main();

// variables
SDL_Surface *screen;
static int g_image_width, g_image_height;
static hotspot *hotspot_last=NULL;

// Macros
#define VERBOSE(...) do { if (cfg_verbose) { printf(__VA_ARGS__); } } while(0)

void printUsage()
{
	printf("Usage: ./laserspotcam [options] config_file\n\n");
	printf("Options:\n");
	printf("  -v         Verbose mode\n");
	printf("  -h         Display this\n");
	printf("  -d         Set v4l device to use. Default: /dev/video0\n");
	printf("  -V         Visual mode. This opens an SDL window you can\n");
	printf("             use to see what the camera sees.\n");
	printf("  -r         Display raw images from the webcam instead\n");
	printf("             of displaying the processed image(threshold applied).\n");
	printf("  -s         Set the number of frames to skip/ignore when a\n");
	printf("             light switch is detected/guessed. Default: 8\n");
	printf("  -b         Detach to background after startup.\n");
	printf("  -t         Test mode (dont run commands)\n");
}

int main(int argc, char **argv)
{
	int res;
	int i;
	int num_hotspots=0;

	printf("laserspotcam version %s, by Raphael Assenat\n", VERSION_STRING);
	
	while((res=getopt(argc, argv, "vhd:Vrs:bt"))>=0)
	{
		switch(res)
		{
			case 'v': cfg_verbose = 1; break;
			case 'h': printUsage(); return 0;
			case 'd': cfg_v4l_dev = optarg; break;
			case 'V': cfg_visual_mode = 1; break;
			case 'r': cfg_display_raw = 1; break;
			case 's': cfg_light_switch_skip = atoi(optarg); break;
			case 'b': cfg_bg_mode = 1; break;
			case 't': cfg_test_mode = 1; break;
			case '?':
			default:
				fprintf(stderr, "Unrecognized option. Try -h for a list of supported options.\n");
				return -1;
		}
	}

	if (argc-optind<1) {
		printf("No config file specified. Starting with -vVr options!\n");
		cfg_visual_mode = 1;
		cfg_display_raw = 1;
		cfg_verbose = 1;
	}
	else
	{
		for (i=optind; i<argc; i++) {
			VERBOSE("Loading config file %s\n", argv[i]);
			res = hotspot_load(argv[i]);
			if (res<0) { 
				fprintf(stderr, "Aborting due to error in config file\n");
				return -1; 
			}
			num_hotspots += res;
		}
	}
	VERBOSE("%d hotspots loaded\n", num_hotspots);
	
	VERBOSE("Ignoring %d frames on light switch events\n", cfg_light_switch_skip);

	


	VERBOSE("Init webcam\n");
	res = initialise_webcam(cfg_v4l_dev);
	if (res<0) {
		fprintf(stderr, "Could not init webcam. Aborting\n");
		return res;
	}


	get_webcam_resolution(&g_image_width, &g_image_height);
	VERBOSE("Webcam image is %d by %d pixels\n", g_image_width, g_image_height);
	
	if (AvgDiff_init(g_image_width,g_image_height)<0) {
		return -1;
	}

	hotspot_setVerboseSystem(cfg_verbose);
	
	if (cfg_visual_mode) {
		return visual_main();
	}

	if (cfg_bg_mode) {
		// at this point, everything is ready. Seems to be a good
		// place to enter daemon mode.
		res = daemonize();
		if (res<0) {
			fprintf(stderr, "Failed to enter background mode. Exiting.\n");
			return -1;
		}
		if (res == 0) {
			// we're the parent. Lets exit.
			return 0;
		}
	}
	
	return daemon_main();
}

int visual_main()
{
	int res, frames_to_skip = 0;
	SDL_Rect motion_rect;
	int motion, idle=1;
	SDL_Surface *wimage;
	char *wimage_color;
	
	VERBOSE("Init SDL\n");
	res = SDL_Init(SDL_INIT_VIDEO);
	if (res<0) {
		fprintf(stderr, "Could not init SDL video\n");
		close_webcam();
		return -1;
	}
	
	screen = SDL_SetVideoMode(g_image_width, g_image_height, 32, SDL_SWSURFACE);
	SDL_WM_SetCaption("laserspotcam", NULL);
	SDL_EnableKeyRepeat(120, 0);

	wimage_color = malloc(g_image_width*g_image_height*3);
	
	// create a surface for the webcam's image
	wimage = SDL_CreateRGBSurfaceFrom(
				wimage_color,
				g_image_width, g_image_height,
				24, g_image_width*3,
				0xff0000, 0xff00, 0xff, 0x00);


	while(1)
	{
		SDL_Event ev;

		while(SDL_PollEvent(&ev))
		{
			switch (ev.type)
			{
				case SDL_QUIT: 
						goto bye;
				case SDL_MOUSEBUTTONDOWN:
						printf("Click at %d,%d\n", ev.button.x, ev.button.y);
						break;
			}
		}

		SDL_v4l_fetchImage();
	
	
		if (cfg_display_raw) {
			tripleBuf(SDL_v4l_get_rgb_buffer(), wimage_color, g_image_width*g_image_height);
		}
		
		AvgDiff_process(SDL_v4l_get_rgb_buffer(), g_image_width, g_image_height);
		filter_threshold(SDL_v4l_get_rgb_buffer(), g_image_width, g_image_height, 64);

		if (frames_to_skip>0)
		{
			VERBOSE("Ignoring frame\n");
			frames_to_skip--;
		}
		else
		{
			motion = calculateMotionBox(SDL_v4l_get_rgb_buffer(), g_image_width, g_image_height, &motion_rect);
			
			if (motion>0) { 
				VERBOSE("Motion detected. %d,%d %dx%d\n",
						motion_rect.x, motion_rect.y,
						motion_rect.w, motion_rect.h);
				if (idle) {			
					if (motion_rect.w * motion_rect.h > 1000) {
						VERBOSE("Too much motion right now\n");
						if (motion_rect.w * motion_rect.h > 
								(g_image_width * g_image_height)/2) {
							VERBOSE("Light switch? Ignoring %d frames...\n", cfg_light_switch_skip);
							frames_to_skip = cfg_light_switch_skip;
						}
					}
					else {
						if (findAndTriggerHotspot(&motion_rect)) {
							VERBOSE("Hotspot activated, not idle\n");
							idle = 0;
						}
					}
				} else {
					VERBOSE("Ignoring motion\n");
				}
			} else {
				if (!idle) {
					VERBOSE("Became idle\n");
				}
				idle = 1;
			}

		}	
		if (!cfg_display_raw) {
			tripleBuf(SDL_v4l_get_rgb_buffer(), wimage_color, g_image_width*g_image_height);
		}

		SDL_BlitSurface(wimage, NULL, screen, NULL);
		draw_hotspots();

		SDL_UpdateRect(screen, 0,0,0,0);
	}

bye:
	SDL_Quit();
	close_webcam();
	return 0;

}

int daemon_main()
{
	SDL_Rect motion_rect;
	int motion, idle=1;
	int frames_to_skip=0;

	while(1)
	{
		SDL_v4l_fetchImage();

		AvgDiff_process(SDL_v4l_get_rgb_buffer(), g_image_width, g_image_height);
		filter_threshold(SDL_v4l_get_rgb_buffer(), g_image_width, g_image_height, 64);

		if (frames_to_skip>0) {
			VERBOSE("Ignoring frame\n");
			frames_to_skip--;
		}
		else
		{
			motion = calculateMotionBox(SDL_v4l_get_rgb_buffer(), g_image_width, g_image_height, &motion_rect);

			if (motion>0) { 
				VERBOSE("Motion detected. %d,%d %dx%d\n",
						motion_rect.x, motion_rect.y,
						motion_rect.w, motion_rect.h);
				if (idle) {			
					if (motion_rect.w * motion_rect.h > 1000) {
						VERBOSE("Too much motion right now\n");
						if (motion_rect.w * motion_rect.h > 
								(g_image_width * g_image_height)/2) {
							VERBOSE("Light switch? Ignoring %d frames...\n", cfg_light_switch_skip);
							frames_to_skip = cfg_light_switch_skip;
						}

					}
					else {
						if (findAndTriggerHotspot(&motion_rect)) {
							VERBOSE("Hotspot activated, not idle\n");
							idle = 0;
						}
					}
				} else {
					VERBOSE("Ignoring motion\n");
				}
			} else {
				if (!idle) {
					VERBOSE("Became idle\n");
				}
				idle = 1;
			}
		}
	}
}


int findAndTriggerHotspot(SDL_Rect *mrect)
{
	hotspot *hs;

	hs = hotspot_getHead();

	while (hs) {

		if (	(mrect->x > hs->x1) && (mrect->y > hs->y1) &&
				(mrect->x + mrect->w <= hs->x2) &&
				(mrect->y + mrect->h <= hs->y2) )
		{
			hotspot_last = hs;
			return hotspot_trig(cfg_test_mode, hs,
					mrect->x + mrect->w/2,
					mrect->y + mrect->h/2);
		}
		
		hs = hs->next;
	}
	hotspot_last = NULL;
	return 0;
}

int calculateMotionBox(char *buf, int w, int h, SDL_Rect *r)
{
	char *first, *last;
	int len;
	int x, X;

	len = (w*h);

	first = memchr(buf, 0xff, len);
	if (!first) { return 0; } // all black
	
	last = memrchr(buf, 0xff, len);
	if (!last) { return 0; } // huh?

	r->y = (int)(first-buf)/(w);
	r->h = (int)(last-buf)/(w) - r->y;

	x = ((int)(first-buf) % (w));
	X = ((int)(last-buf) % (w) );

		
	r->x = x < X ? x : X;
	r->w = x < X ? (X-x+1) : (x-X+1);
	
	return 1;
}

void waitKeyPress()
{
	SDL_Event event;

	while (1)
	{
		SDL_WaitEvent(&event);

		if (event.type==SDL_KEYDOWN)
		{
			return;
		}
	}
}


/**
 * Draws a grid with specified width and height for each
 * cell.
 *
 * Uses the 'screen' global SDL surface.
 */
void draw_grid(int w, int h)
{
	SDL_Rect r;
	Uint32 color = SDL_MapRGB(screen->format, 0xff, 0x00, 0x00);
	int x,y;
	
	r.x = 0; r.w = screen->w; r.h = 1;
	for (y=0; y<screen->h; y+=h) {
		r.y = y;
		SDL_FillRect(screen, &r, color);
	}

	r.y = 0; r.h = screen->h; r.w = 1;
	for (x=0; x<screen->w; x+=w) {
		r.x = x;
		SDL_FillRect(screen, &r, color);
	}
}

void draw_hotspots(void)
{
	hotspot *hs;
	Uint32 color = SDL_MapRGB(screen->format, 0xff, 0x00, 0x00);
	Uint32 color_active = SDL_MapRGB(screen->format, 0x00, 0xff, 0x00);

	hs = hotspot_getHead();

	while (hs) {
		if (hs == hotspot_last) {
			draw_box(color_active, 
					hs->x1, hs->y1, 
					hs->x2 - hs->x1,
					hs->y2 - hs->y1);
		}
		else {
			draw_box(color, hs->x1, hs->y1, 
				hs->x2 - hs->x1,
				hs->y2 - hs->y1);
		}
		
		hs = hs->next;
	}
}

void draw_box(Uint32 color, int x, int y, int w, int h)
{
	SDL_Rect rect;

	// top
	rect.x = x; rect.y = y;
	rect.w = w; rect.h = 1;
	SDL_FillRect(screen, &rect, color);

	// left
	rect.x = x; rect.y = y;
	rect.w = 1; rect.h = h;
	SDL_FillRect(screen, &rect, color);

	// right
	rect.x = x + w;
	rect.w = 1; 
	rect.h = h; rect.y = y;
	SDL_FillRect(screen, &rect, color);

	// bottom
	rect.x = x;
	rect.w = w;
	rect.h = 1;
	rect.y = y + h;
	SDL_FillRect(screen, &rect, color);
}

void draw_cross(void)
{
	int w, h, y, x, p;

	if ( SDL_MUSTLOCK(screen) ) {
		if ( SDL_LockSurface(screen) < 0 ) {
			return;
		}
	}

	w = screen->w;
	h = screen->h;
	x = screen->w/2; // where to put vertical line
	y = screen->h/2; // where to put horizontal line

	/* draw the vertical line */
	for (p=0; p<h; p++) {
		putpixel(screen, x, p, SDL_MapRGB(screen->format, 255, 255, 0));
	}

	/* draw the horizontal line */
	for (p=0; p<w; p++) {
		putpixel(screen, p, y, SDL_MapRGB(screen->format, 255, 255, 0));
	}


	if ( SDL_MUSTLOCK(screen) ) {
		SDL_UnlockSurface(screen);
	}
}

// convert a grayscale buffer to a 24 bit, rgb buffer
void tripleBuf(unsigned char *src, unsigned char *dst, int srclen)
{
	do
	{
		*dst = *src; dst++;
		*dst = *src; dst++;
		*dst = *src; dst++;
		//*dst = 0xff;

		src++;		
	} while (--srclen);
}


