/*
laserspotcam        A software to track a laser pointer on a wall.
Copyright (C) 2005  Raphael Assenat

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Author: Raphael Assenat <raph@raphnet.net>
*/
/* Functions that controls the webcam, does ioctls, change settings... */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "SDL.h"

#include "webcam.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* Video for linux and philips webcam includes */
#include <linux/videodev.h>
#include <sys/ioctl.h>
#include "pwc-ioctl.h"


static int g_webcam_descriptor=-1;

static void print_picture_properties(struct video_picture videopic);
static int get_picture_properties(struct video_picture *videopic);
static int get_video_window(struct video_window *videowin);

static int current_capture_width;
static int current_capture_height;
static int changed_capture_size=1;

#define DEFAULT_CAPTURE_WIDTH 320
#define DEFAULT_CAPTURE_HEIGHT 240
#define DEFAULT_CAPTURE_RATE 10

int initialise_webcam(char *device_file)
{
	struct video_window videowin; /* To play with the video window. eg: change size */
	struct video_picture videopic; /* To set the capture format to YUV4:2:0P*/

	/* Open device */
	g_webcam_descriptor = open(device_file, O_RDONLY);
	if (g_webcam_descriptor<0) {
		perror("open");
		return -1;
	}

	/* Read the current webcam settings */
	get_picture_properties(&videopic);
	get_video_window(&videowin);

	/* Set the capture size */
	videowin.width=DEFAULT_CAPTURE_WIDTH;
	videowin.height=DEFAULT_CAPTURE_HEIGHT;
	ioctl(g_webcam_descriptor, VIDIOCSWIN, &videowin);

	/* Get the capture size from the cam */
	get_video_window(&videowin);
	current_capture_width=videowin.width;
	current_capture_height=videowin.height;

	/* Set the capture palette */
	ioctl(g_webcam_descriptor, VIDIOCGPICT, &videopic);
	videopic.palette = VIDEO_PALETTE_YUV420P;
	videopic.depth = 12;
	ioctl(g_webcam_descriptor, VIDIOCSPICT, &videopic);

	/* Set the frame rate */
	//set_webcam_capture_rate(DEFAULT_CAPTURE_RATE);

	//set_webcam_white_balance_mode(0);
	set_webcam_agc(-1);
	//set_webcam_shutter_speed(-1);
	//set_webcam_led_cycle(0, 10);
	//set_webcam_compression_preference(0);
	//printf("Compression level: %d\n", get_webcam_compression_preference() );
	is_webcam_philips();
	//set_webcam_gamma_correction(32000);
	//set_webcam_colour_saturation(65535);
	//print_picture_properties(videopic);

	return 0;
}

int close_webcam()
{
	close(g_webcam_descriptor);
	return 0;
}

unsigned short set_webcam_brightness(unsigned short brightness)
{
	int res;
	struct video_picture videopic;
	printf("trying to set it to: %d\n", brightness);
	/* Get current settings */
	get_picture_properties(&videopic);
	printf("Before: %d\n", videopic.brightness);
	/* Apply settings */
	videopic.brightness = brightness;
	res = ioctl(g_webcam_descriptor, VIDIOCSPICT, &videopic);
	printf("res: %d\n", res);
	printf("so: %d\n", videopic.brightness);
	/* Return the effective setting */
	get_picture_properties(&videopic);
	printf("After: %d\n", videopic.brightness);
	return videopic.brightness;
}

unsigned short get_webcam_brightness()
{
	struct video_picture videopic;
	/* Return the effective setting */
	get_picture_properties(&videopic);
	return videopic.brightness;
}

/* */

unsigned short set_webcam_contrast(unsigned short contrast)
{
	struct video_picture videopic;

	/* Get current settings */
	get_picture_properties(&videopic);

	/* Apply settings */
	videopic.contrast = contrast;
	ioctl(g_webcam_descriptor, VIDIOCSPICT, &videopic);

	/* Return the effective setting */
	get_picture_properties(&videopic);
	return videopic.contrast;
}

unsigned short get_webcam_contrast()
{
	struct video_picture videopic;
	/* Return the effective setting */
	get_picture_properties(&videopic);
	return videopic.contrast;
}

unsigned short set_webcam_gamma_correction(unsigned short gamma)
{
	struct video_picture videopic;

	/* Get current settings */
	get_picture_properties(&videopic);

	/* Apply settings */
	videopic.whiteness = gamma;
	ioctl(g_webcam_descriptor, VIDIOCSPICT, &videopic);

	/* Return the effective setting */
	get_picture_properties(&videopic);
	return videopic.whiteness;
}

unsigned short get_webcam_gamma_correction()
{
	struct video_picture videopic;
	/* Return the effective setting */
	get_picture_properties(&videopic);
	return videopic.whiteness;
}

unsigned short set_webcam_colour_saturation(unsigned short saturation)
{
	struct video_picture videopic;

	/* Get current settings */
	get_picture_properties(&videopic);

	/* Apply settings */
	videopic.colour = saturation;
	ioctl(g_webcam_descriptor, VIDIOCSPICT, &videopic);

	/* Return the effective setting */
	get_picture_properties(&videopic);
	return videopic.colour;
}

unsigned short get_webcam_colour_saturation()
{
	struct video_picture videopic;
	/* Return the effective setting */
	get_picture_properties(&videopic);
	return videopic.colour;
}



char is_webcam_philips()
{
	struct video_capability vcap;
	int type;

	if (ioctl(g_webcam_descriptor, VIDIOCGCAP, &vcap) < 0) { return 0; }

	if (sscanf(vcap.name, "Philips %d webcam", &type) < 1) {
//		printf("No Philips webcam detected.\n");
		return 0;
	}
	else {
//		printf("Philips webcam type %d detected.\n", type);
		return 1;
	}
	return 0; /* Not reached, but prevents warnings ;) */
}

int get_webcam_name(char *name, int len)
{
	struct video_capability vcap;

	if (ioctl(g_webcam_descriptor, VIDIOCGCAP, &vcap) < 0) { return 0; }

	strncpy(name, vcap.name, len);
	name[len-1]=0;

	return 0; /* Not reached, but prevents warnings ;) */
}

int set_webcam_compression_preference(int level)
{
	if (level<=0) { level=0; }
	if (level>3) { level=3; }
	ioctl(g_webcam_descriptor, VIDIOCPWCSCQUAL, &level);
	return level;
}

int get_webcam_compression_preference(void)
{
	int level;
	ioctl(g_webcam_descriptor, VIDIOCPWCGCQUAL, &level);
	return level;
}

void restore_webcam_factory_default(void)
{
	ioctl(g_webcam_descriptor, VIDIOCPWCFACTORY, NULL);
}

void store_settings_in_webcam(void)
{
	ioctl(g_webcam_descriptor, VIDIOCPWCSUSER, NULL);
}

void restore_settings_from_webcam(void)
{
	ioctl(g_webcam_descriptor, VIDIOCPWCRUSER, NULL);
}

int set_webcam_led_cycle(int ms_on, int ms_off)
{
	struct pwc_leds LED;
	LED.led_on = ms_on;
	LED.led_off = ms_off;

	ioctl(g_webcam_descriptor, VIDIOCPWCSLED, &LED);

	return 0;
}


int set_webcam_shutter_speed(int speed)
{
	ioctl(g_webcam_descriptor, VIDIOCPWCSSHUTTER, speed); // Set the agc

	return speed;
}

int set_webcam_agc(int agc)
{
	/* Philips webcam driver specific function */
	/* Give a number between 0 and 65535 to set the Gain manually,
	 * Give a negative number to enable automatic Gain control */

	/* Set the webcam at specific agc value */
	ioctl(g_webcam_descriptor, VIDIOCPWCSAGC, &agc); // Set the agc

	/* Get the webcam real agc value */
	ioctl(g_webcam_descriptor, VIDIOCPWCGAGC, &agc);
	return agc;
}

int get_webcam_agc()
{
	int agc;
	ioctl(g_webcam_descriptor, VIDIOCPWCGAGC, &agc);
	return agc;
}

int set_webcam_white_balance_mode(int mode)
{
	/* Philips webcam driver specific function */
	struct pwc_whitebalance white_balance;

	if (ioctl(g_webcam_descriptor, VIDIOCPWCGAWB, &white_balance) < 0){
		perror("VIDIOCPWCGAWB");
		close(g_webcam_descriptor);
		exit(1);
	}

	switch(mode)
	{
		case 0:	white_balance.mode = PWC_WB_AUTO; break;
		case 1:	white_balance.mode = PWC_WB_MANUAL; break;
		case 2:	white_balance.mode = PWC_WB_INDOOR; break;
		case 3:	white_balance.mode = PWC_WB_OUTDOOR; break;
		case 4:	white_balance.mode = PWC_WB_FL; break;
		default:	white_balance.mode = PWC_WB_AUTO; break;
	}

	if (ioctl(g_webcam_descriptor, VIDIOCPWCSAWB, &white_balance) < 0){
		perror("VIDIOCPWCSAWB");
		close(g_webcam_descriptor);
		exit(1);
	}
	return 0;
}

int get_webcam_white_balance_mode()
{
	struct pwc_whitebalance white_balance;

	if (ioctl(g_webcam_descriptor, VIDIOCPWCGAWB, &white_balance) < 0){
		perror("VIDIOCPWCGAWB");
		close(g_webcam_descriptor);
		exit(1);
	}
	//printf("%d\n", white_balance.mode);
	switch(white_balance.mode)
	{
		case PWC_WB_AUTO: return 0; break;
		case PWC_WB_MANUAL: return 1; break;
		case PWC_WB_INDOOR: return 2; break;
		case PWC_WB_OUTDOOR: return 3; break;
		case PWC_WB_FL: return 4; break;
	}

	return -1;
}

int get_webcam_white_balance_values(int *r, int *b)
{
	struct pwc_whitebalance white_balance;

	if (ioctl(g_webcam_descriptor, VIDIOCPWCGAWB, &white_balance) < 0){
		perror("VIDIOCPWCGAWB");
		close(g_webcam_descriptor);
		exit(1);
	}
	*r = white_balance.read_red;
	*b = white_balance.read_blue;
	return 0;
}

int set_webcam_capture_rate(char rate)
{
	struct video_window videowin;

	if (rate>30) { rate=30; }
	if (rate<5) { rate=5; }

	/* Get the current settings */
	get_video_window(&videowin);

	/* Set the new frame rate */
	videowin.flags &= ~PWC_FPS_FRMASK;
	videowin.flags |= (rate << PWC_FPS_SHIFT);

	/* Commit changes */
	ioctl(g_webcam_descriptor, VIDIOCSWIN, &videowin);

	/* Check the new settings(if any) */
	get_video_window(&videowin);

	/* Return the new frame rate */
	return (videowin.flags & PWC_FPS_FRMASK) >> PWC_FPS_SHIFT;
}

int get_webcam_capture_rate()
{
	struct video_window videowin;

	/* Check the settings */
	get_video_window(&videowin);

	/* Return the frame rate */
	return (videowin.flags & PWC_FPS_FRMASK) >> PWC_FPS_SHIFT;
}

int set_webcam_resolution(int w, int h)
{
	struct video_window videowin; /* To play with the video window. eg: change size */
	get_video_window(&videowin);

	/* Set the resolution */
	videowin.width=w;
	videowin.height=h;
	ioctl(g_webcam_descriptor, VIDIOCSWIN, &videowin);

	changed_capture_size=1;

	current_capture_height=h;
	current_capture_width=w;

	return 0;
}

int get_webcam_resolution(int *w, int *h)
{
	struct video_window videowin; /* To play with the video window. eg: change size */

	if (changed_capture_size)
	{
		get_video_window(&videowin);
		*w = videowin.width;
		*h = videowin.height;
		changed_capture_size=0;
		current_capture_height=*h;
		current_capture_width=*w;
		return 0;
	}
	*w = current_capture_width;
	*h = current_capture_height;
	return 0;
}

int get_webcam_filedescriptor()
{
	return g_webcam_descriptor;
}


/* Used by webcam.c */
int get_picture_properties(struct video_picture *videopic)
{
	if (ioctl(g_webcam_descriptor, VIDIOCGPICT, videopic) < 0)
	{
		perror("VIDIOCGPICT");
		close(g_webcam_descriptor);
		exit(1);
	}
	return 1;
}

int get_video_window(struct video_window *videowin)
{
	if (ioctl(g_webcam_descriptor, VIDIOCGWIN, videowin) < 0)
	{
		perror("VIDIOCGWIN");
		close(g_webcam_descriptor);
		exit(1);
	}
	return 1;
}

void print_picture_properties(struct video_picture videopic)
{
	printf("\nGet picture properties:\n");
	printf("Brightness: %d\n",videopic.brightness);
	printf("       Hue: %d\n",videopic.hue);
	printf("    Colour: %d\n",videopic.colour);
	printf("  Contrast: %d\n",videopic.contrast);
	printf(" Whiteness: %d\n",videopic.whiteness);
	printf("     Depth: %d\n",videopic.depth);
	printf("   Palette: %d\n",videopic.palette);
}
